module LPPE where

import Processes
import DataSpec
import Expressions
import Auxiliary
import List
import Debug.Trace

type System        = (LPPE, InitialState)
type Specification = (LPPE, InitialState, DataSpec)
type Constant      = String
type InitialState  = [Constant]
type CommFunction  = Action -> Action -> Action

data LPPE = LPPE ProcessName ProcessPars [Summand] deriving (Show, Eq)
	
data LPPEShowPRCRL = LPPEShowPRCRL Specification
instance Show LPPEShowPRCRL where
  show (LPPEShowPRCRL (LPPE name params summands, initialState, dataspec)) = datastring ++ "\n"  ++ name ++ printPairs (newParams) ++ 
                                                      " = \n      " ++ summandsString ++ "\n\ninit "++ name ++ "[" ++ (infixString initialState ", ") ++ "]"
    where
      summandsString = infixString (map (printSummandPRCRL dataspec name params) summands) " \n    ++ "
      types                 = nub (["Bool"] 
                            ++ map (printType.snd) params ++ [printType t | summand <- summands, (v,t) <- getLocalPars summand]
                            ++ [printType t | summand <- summands, (v,t,f) <- getProbChoices summand] )
      functions      = concat (map getFunctionsInSummand summands)
      dataspec2      = removeUnusedTypesAndFunctions dataspec types functions
      datastring     = "\n  \n" ++ printDataSpec dataspec2 ++ "\n  \n"
      newParams      = map (\x -> (fst x, printType (snd x))) params

data LPPEShow = LPPEShow Specification Bool
instance Show LPPEShow where
  show (LPPEShow (LPPE name params summands, initialState, dataspec) style) = name ++ printPairs newParams ++ " = \n      " ++ summandsString
                                                                              ++ "\n\nInitial state: " ++ name ++ commaList initialState
    where
      summandsString = infixString (map (printSummand dataspec style name params) summands) " \n    + "
      newParams      = map (\x -> (fst x, printType (snd x))) params
      
type LocalPars   = [(Variable, Type)]
type Condition   = Expression
type ProbChoice = (Variable, Type, ProbDistr)
type Summand     = (LocalPars, Condition, Action, ActionPars, [ProbChoice], NextPars)

removeUnusedTypesAndFunctions :: DataSpec -> [String] -> [String] -> DataSpec
removeUnusedTypesAndFunctions dataspec types functions 
  = ([t | t <- fst3 dataspec, elem (getTypeName t) types], 
     [(TypeName t,v) | (TypeName t,v) <- snd3 dataspec, elem t types], 
     [FunctionDef f mappings | FunctionDef f mappings <- thd3 dataspec, elem f functions])

printType (TypeName name)     = name
printType (TypeRange from to) = "{" ++ show from ++ ".." ++ show to ++ "}"


------------------------------------
-- Functions for displaying LPPEs --
------------------------------------

-- This function prints a summand in pretty-print style.
printSummand :: DataSpec -> Bool -> ProcessName -> ProcessPars -> Summand -> String
printSummand dataspec style procName procPars (params, c, a, aps, probChoices, g) = 
  "(" ++ localParsToString ++ show c ++ " => " ++ a ++ commaList (map show aps)  
  ++ " . " ++ printNextState dataspec style procName procPars (params, c, a, aps, probChoices, g) probChoices g
  ++ (if (length params == 0) then "" else ")") ++ ")"
    where
      parsList = (infixString [left ++ " : " ++ (printType right) | (left, right) <- params] ", ")
      localParsToString | length params == 0 = ""
	                    | otherwise          = "sum(" ++ parsList ++ ", "

-- This function prints a summand in prCRL style.
printSummandPRCRL :: DataSpec -> ProcessName -> ProcessPars -> Summand -> String
printSummandPRCRL dataspec procName procPars (params, c, a, aps, probChoices, g) = 
  "(" ++ localParsToString ++ printExpression c ++ " => " ++ a ++ commaList (map (printExpression) aps)  
  ++ " . " ++ printNextState dataspec True procName procPars (params, c, a, aps, probChoices, g) probChoices g
  ++ (if (length params == 0) then "" else ")") ++ ")"
    where
      parsList = (infixString [left ++ " : " ++ (printType right) | (left, right) <- params] ", ")
      localParsToString | length params == 0 = ""
	                    | otherwise          = "sum(" ++ parsList ++ ", "

-- This function prints an expression in prCRL-style
printExpression :: Expression -> String
printExpression (Variable variable)   = variable
printExpression (Function name pars)  = name ++ optionalParentheses (infixString (map printExpression pars) ", ")

-- This function prints the next state	
printNextState :: DataSpec -> Bool -> ProcessName -> ProcessPars -> Summand -> [ProbChoice] -> NextPars -> String
printNextState dataspec style procName procPars summand probChoices g 
  | easier    = procName ++ printUpdates style procPars summand newNextState
  | otherwise = "psum(" ++ (probChoicesString) ++ " : " ++ procName ++ printUpdates style procPars summand g ++ ")"
  where
    easier            = and (map (== Variable "1") (map thd3 probChoices))
    substitutions     = [(v,Variable ((allValues t (fst3 dataspec))!!0)) | (v,t,Variable "1") <- probChoices]
    newNextState      = map (substituteInExpression substitutions) g
    probChoicesString = infixString (map (\x -> fst3 x ++ ": " ++ printType (snd3 x) ++ ", " ++ show (thd3 x)) probChoices) " || "

-- This function prints the updates to the state
printUpdates :: Bool -> ProcessPars -> Summand -> NextPars -> String
printUpdates style procPars (params, c, a, aps, probChoices, g) next = "[" ++ (infixString (getUpdates style params probChoices (zip (map fst procPars) next)) ", ") ++ "]"

-- This function produces the updates to the state
getUpdates style params probChoices [] = []
getUpdates True params probChoices ((parName, nextState):xs) = (show nextState):(getUpdates True params probChoices xs)
getUpdates False params probChoices ((parName, nextState):xs) | unchanged = getUpdates False params probChoices xs
                                                              | otherwise = (parName ++ " := " ++ show nextState):(getUpdates False params probChoices xs)
  where
	unchanged = nextState == Variable parName && (not (elem parName (map fst params)) ) && (not(elem parName (map fst3 probChoices)))

printInit :: [String] -> String
printInit x = "[" ++ infixString x "," ++ "]"

-----------------------------------------
-- General functions for dealing with  --
-- LPPEs, Specifications, and Summands --
----------------------------------------

-- Given a specification, provides the LPPE
getLPPE :: Specification -> LPPE
getLPPE (lpe, initialState, dataspec) = lpe

-- Given a specification, provides the initial state
getInitialState :: Specification -> InitialState
getInitialState (lpe, initialState, dataspec) = initialState

-- Given a specification, provides the data part
getDataSpec :: Specification -> DataSpec
getDataSpec (lpe, initialState, dataspec) = dataspec

-- Provides the global parameters of an LPPE.
getLPPEPars :: LPPE -> ProcessPars
getLPPEPars (LPPE name pars summands) = pars

-- Looks up a summand of an LPPE.
getSummand :: LPPE -> Int -> Summand
getSummand (LPPE name pars summands) i = summands!!i

-- Provideds all summands of an LPPE.
getSummands :: LPPE -> [Summand]
getSummands (LPPE name pars summands) = summands

-- Provides a list of summand indices
getSummandNrs :: LPPE -> [Int]
getSummandNrs lppe = [0..length (getSummands lppe) - 1]

-- Provides the action of a summand.
getAction :: Summand -> Action
getAction (params, c, a, aps, probChoices, g) = a

-- Changes the action of a summand.
setAction :: Summand -> Action -> Summand
setAction (params, c, a, aps, probChoices, g) aNew = (params, c, aNew, aps, probChoices, g)

-- Provides all actions of an LPPE
getActions :: LPPE -> [Action]
getActions lppe = map getAction (getSummands lppe)

-- Provides the condition of a summand.
getCondition :: Summand -> Expression
getCondition (params, c, a, aps, probChoices, g) = c

-- Provides the action of a summand
getActionPars :: Summand -> ActionPars
getActionPars (params, c, a, aps, probChoices, g) = aps

-- Provides the local parameters of a summand
getLocalPars :: Summand -> LocalPars
getLocalPars (params, c, a, aps, probChoices, g) = params

-- Provides the probabililstic summations of a summand
getProbChoices :: Summand -> [ProbChoice]
getProbChoices (params, c, a, aps, probChoices, g) = probChoices


-- Provides the next state for a certain parameter, given a summand.
getNextState :: Summand -> Int -> Expression
getNextState (params, c, a, aps, probChoices, g) i = g!!i

-- This function takes a summand and adds a conjunct to its condition.
addImplication :: Expression -> Summand -> Summand
addImplication cNew (params, c, a, aps, probChoices, g) = (params, Function "and" [c, cNew] , a, aps, probChoices, g)

-- This function takes an LPPE, an expression and a list of summands indices,
-- and adds the expression as a condition to every of these summands
addCondition :: LPPE -> Expression -> [Int] -> LPPE
addCondition (LPPE name pars summands) c nrs = LPPE name pars [if elem i nrs then addImplication c (summands!!i) else summands!!i | i <- [0..length summands - 1]]

-- This function takes a summand and adds a summation to its list of local parameters.
addSummation :: LocalPars -> Summand -> Summand
addSummation localPars (params, c, a, aps, probChoices, g) = (params ++ localPars, c, a, aps, probChoices, g)

-- This function substitutes values for parameters in an LPPE.
substituteInLPPE :: LPPE -> [(Int, String)] -> LPPE
substituteInLPPE (LPPE name pars summands) substitutions = LPPE name pars (map (substituteInSummand subs) summands)
  where
    subs = [(fst (pars!!parNr), Variable value) | (parNr, value) <- substitutions]

-- This function substitutes values for parameters in a summand.
substituteInSummand :: [Substitution] -> Summand -> Summand
substituteInSummand subs (params, c, a, aps, probChoices, g) = (params, newCondition, a, newAps, newProbChoices, newG)
  where
    subs2          = removeFromSubstitutionsList (map fst params) subs
    newCondition   = substituteInExpression subs2 c
    newAps         = map (substituteInExpression subs2) aps
    subs3          = removeFromSubstitutionsList (map fst3 probChoices) subs2
    newProbChoices = map (\x -> (fst3 x, snd3 x, substituteInExpression subs3 (thd3 x))) probChoices
    newG           = map (substituteInExpression subs3) g

getFunctionsInSummand :: Summand -> [String]
getFunctionsInSummand (params, c, a, aps, probChoices, g) = result
  where
    result = functionsInExpression c ++ concat (map functionsInExpression aps) ++ 
             concat (map functionsInExpression (map thd3 probChoices)) ++
             concat (map functionsInExpression g)

-- This function encapsulates (removes) actions in a system.
encapsulate :: System -> [Action] -> System
encapsulate system []                                        = system
encapsulate (LPPE name pars summands, init) (action:actions) = encapsulate newSystem actions
  where
    newSystem = (LPPE name pars [s | s <- summands, getAction s /= action], init)

-- This function hides actions in a system.
hide :: System -> [Action] -> System
hide system []                                        = system
hide (LPPE name pars summands, init) (action:actions) = hide newSystem actions
  where
    newSystem = (LPPE name pars (   [(params, c, "tau", [], probChoices, g) | (params, c, a, aps, probChoices, g) <- summands , a == action]
                                 ++ [s | s <- summands , getAction s /= action]), init)

-- This function renames actions in a system.
rename :: System -> [(Action, Action)] -> System
rename system []                                        = system
rename (LPPE name pars summands, init) ((actFrom, actTo):actions) = rename newSystem actions
  where
    newSystem = (LPPE name pars (   [(params, c, actTo, [], probChoices, g) | (params, c, a, aps, probChoices, g) <- summands , a == actFrom]
                                 ++ [s | s <- summands , getAction s /= actFrom]), init)


-- This function removes (constant) parameters from an LPPE.
removeParametersFromLPPE :: Specification -> [Int] -> Specification
removeParametersFromLPPE ((LPPE name pars summands),initial,dataspec) indices = (LPPE name newPars newSummands, newInitial, dataspec)
  where
    indicesToKeep = [i | i <- [0..length pars - 1], not(elem i indices)]
    newPars       = [pars!!i | i <- indicesToKeep]
    newSummands   = map (removeParametersFromSummand indicesToKeep) summands
    newInitial    = [initial!!i | i <- [0..length initial - 1], not(elem i indices)]

-- This function changes the next state function of a summand by omitting
-- some of the parameters. The parameter of this function indicates which
-- parameters should be kept.
removeParametersFromSummand indicesToKeep (params, c, a, aps, probChoices, g) = (params, c, a, aps, probChoices, nextStatePars)
  where
    nextStatePars = [g!!i | i <- indicesToKeep]

---------------------------------------
-- Function for parallel composition --
---------------------------------------

-- This function suffixes all variables of an LPPE by a string.
suffixStringLPPE suffix (LPPE name pars summands) = LPPE name newPars newSummands
  where
    newPars       = [(a ++ suffix,b)  | (a,b) <- pars] 
    substitutions = [(a, a ++ suffix) | (a,b) <- pars]
    newSummands   = map (suffixStringSummand suffix substitutions) summands

-- This function suffixes all variables of a summand by a string.
suffixStringSummand suffix substitutions (params, c, a, aps, [(v, t, f)], g) = (newParams, newC, a, newAps, [(newV, t, newF)], newG)
   where
     newParams = [(a ++ suffix,b) | (a,b) <- params]
     newSubs   = nub (substitutions ++ [(a, a ++ suffix) | (a,b) <- params])
     newC      = substituteInExpression [(v,Variable e) | (v,e) <- newSubs] c
     newAps    = map (substituteInExpression [(v,Variable e) | (v,e) <- newSubs]) aps
     newV      = v ++ suffix
     newSubs2  = nub (newSubs ++ [(v, newV)])
     newF      = substituteInExpression [(v,Variable e) | (v,e) <- newSubs2] f
     newG      = map (substituteInExpression [(v,Variable e) | (v,e) <- newSubs2]) g
suffixStringSummand suffix substitutions (params, c, a, aps, probChoices, g) = error("Initially, all summands should have one probabilistic choice.")

-- This function takes two systems and a communication function,
-- and produces their parallel composition.
composeSystems :: Bool -> System -> System -> (CommFunction) -> (LPPE, [[Char]])
composeSystems prism ((LPPE name1 pars1 summands1),initial1) ((LPPE name2 pars2 summands2),initial2) c
  = (LPPE "Z" (pars1 ++ pars2) (combineSummands prism summands1 summands2 pars1 pars2 c), initial1 ++ initial2)

-- This function takes two lists of summands of different processes,
-- and produces the corresponding summands in the parallel composition
-- of these processes.
combineSummands :: Bool -> [Summand] -> [Summand] -> ProcessPars -> ProcessPars -> CommFunction -> [Summand]
combineSummands prism summands1 summands2 pars1 pars2 c = alone ++ together
  where
    shared   = getSharedActions summands1 summands2
    alone    = mergeSummands (if prism then shared else []) summands1 summands2 pars1 pars2
    comm     = if prism then  \x -> \y -> createCommFunction x y [(a,a,a) | a <- shared, a /= "tau"] else c
    together = communicateSummands [(a,b) | a <- summands1, b <- summands2] comm

-- This function obtains the actions over which two systems (given by their summands)
-- can/must synchronise.
getSharedActions :: [Summand] -> [Summand] -> [Action]
getSharedActions summands1 summands2 = intersect (map getAction summands1) (map getAction summands2)

-- This function takes the union of the summands of two processes,
-- changing their next state function to take into account the new
-- parameters of the product process.
-- The actions provided to this functions are assumed to be shared actions,
-- and are mandatory to synchronise on.
mergeSummands :: [Action] -> [Summand] -> [Summand] -> ProcessPars -> ProcessPars -> [Summand]
mergeSummands sharedActions [] [] pars1 pars2                                            
  = []
mergeSummands sharedActions ((params, c, a, aps, probChoices, g):summands1) summands2 pars1 pars2 
  | a == "tau" || not (elem a sharedActions) = [(params, c, a, aps, probChoices, mergeNextStatesLeft g pars2)]  ++ (mergeSummands sharedActions summands1 summands2 pars1 pars2)
  | otherwise                                = mergeSummands sharedActions summands1 summands2 pars1 pars2
mergeSummands sharedActions [] ((params, c, a, aps, probChoices, g):summands2) pars1 pars2              
  | a == "tau" || not (elem a sharedActions) = [(params, c, a, aps, probChoices, mergeNextStatesRight g pars1)]  ++ (mergeSummands sharedActions [] summands2 pars1 pars2)
  | otherwise                                = mergeSummands sharedActions [] summands2 pars1 pars2

-- This function changes the next state function to take
-- into account the new parameters (leaving them unchanged).
mergeNextStatesLeft  n pars2 = n ++ [Variable x | x <- (map fst pars2)]
mergeNextStatesRight n pars1 = [Variable x | x <- (map fst pars1)] ++ n

-- This function produces the summands that arise in parallel composition
-- by communication of actions.
communicateSummands :: [(Summand, Summand)] -> CommFunction -> [Summand]
communicateSummands [] comm = []
communicateSummands ((summand1, summand2):xs) comm
  | communicable = communicateSummands xs comm ++ merge summand1 summand2 commAction 
  | otherwise        = communicateSummands xs comm
  where
    commAction   = comm (getAction summand1) (getAction summand2)
    communicable = commAction /= "" && length (getActionPars summand1) == length (getActionPars summand2)

-- This function merges two summands that can communicate
-- over an action.
merge :: Summand -> Summand -> Action -> [Summand]
merge (params1, c1, a1, aps1, probChoices1, g1) (params2, c2, a2, aps2, probChoices2, g2) action = result
     where
        apsEqual = [Function "eq" [a,b] | (a,b) <- zip aps1 aps2]
        result   = [(params1 ++ params2, Function "and" ([c1, c2] ++ apsEqual), action, aps1, probChoices1 ++ probChoices2, g1 ++ g2)]

-- This function creates a communication function based on a list of 
-- (action, action,action) pairs. It checks whether the inputs x and y
-- can communicate, and if so, returns the corresponding communication action.
createCommFunction :: Action -> Action -> [(Action, Action, Action)] -> Action
createCommFunction x y []                          = ""
createCommFunction x y ((a,b,c):rest) | match      = c
                                      | otherwise  = createCommFunction x y rest 
  where
    match = (x == a && y == b) || (x == b && y == a)


-------------------------------------------------------------
-- Functions to check whether a process is already in LPPE --
-- form and to parse it indeed to LPPE in that case.       --
-------------------------------------------------------------

-- This function checks whether a process term is already
-- in LPPE format.
isLPPE :: String -> Int -> ProcessTerm -> Bool
isLPPE name nrPars (Plus summands) = and (map (checkSummands name nrPars) summands)
isLPPE name nrPars term            = checkSummations name nrPars term

-- This functions checks whether a process term is a summand
-- by unfoldings the nondeterministic choices.
checkSummands name nrPars (Plus summands) = and (map (checkSummands name nrPars) summands)
checkSummands name nrPars term            = checkSummations name nrPars term

-- This function walks through the summations of the summand.
checkSummations :: String -> Int -> ProcessTerm -> Bool
checkSummations name nrPars (Sum v t rhs)                                           = checkSummations name nrPars rhs
checkSummations name nrPars (Implication c rhs)                                     = checkConditions name nrPars rhs
checkSummations name nrPars (ActionPrefix a aps probs (ProcessInstance name2 pars)) = name == name2 && length pars == nrPars
checkSummations name nrPars other                                                   = False

-- This function walks through the conditions of the summand.
checkConditions :: String -> Int -> ProcessTerm -> Bool
checkConditions name nrPars (Implication c rhs) = checkConditions name nrPars rhs
checkConditions name nrPars (ActionPrefix a aps probs (ProcessInstance name2 pars)) = name == name2 && length pars == nrPars
checkConditions name nrPars other  = False

-- Given that a process term is in the right format, this function indeed
-- parses it to a list of summands.
parseLPPE :: ProcessTerm -> [Summand]
parseLPPE (Plus [a,b]) = (parseLPPE a) ++ (parseLPPE b)
parseLPPE x            = [parseSummand x [] (Variable "T")]

parseSummand :: ProcessTerm -> ProcessPars -> Expression -> Summand
parseSummand (Sum v t rhs) pars c                                       = parseSummand rhs (pars ++ [(v,t)]) c
parseSummand (Implication cNew rhs) pars c | c == (Variable "T")        = parseSummand rhs pars cNew
parseSummand (Implication cNew rhs) pars c | otherwise                  = parseSummand rhs pars (Function "and" [c, cNew])
parseSummand (ActionPrefix a aps probs (ProcessInstance name p)) pars c = (pars, c, a, aps, probs, p)
parseSummand other pars c
  = error("Trying to interpret a specification as an LPPE even though it is not.")

----------------------
-- For PRISM parser --
----------------------

fixGlobalVariables :: System -> System
fixGlobalVariables ((LPPE name pars summands), initial) = (LPPE name newPars newSummands, newInits)
  where
    (globals, newPars, newInits) = findGlobalVariables pars initial
    newSummands        = map (fixGlobalsInSummand globals pars) summands

findGlobalVariables :: ProcessPars -> InitialState -> ([String], ProcessPars, InitialState)
findGlobalVariables [] []                  = ([],[],[])
findGlobalVariables ((v,t):pars) (i:inits) = (global ++ restGlobals, (v,t):restPars, i:restInits)
  where
    newPars                            = [p | p <- pars, p /= (v,t)]
    indices                            = [i | i <- [0..length inits - 1], pars!!i /= (v,t)]
    newInits                           = [inits!!i | i <- indices]
    global                             = if newPars == pars then [] else [v]
    (restGlobals, restPars, restInits) = findGlobalVariables newPars newInits
 
fixGlobalsInSummand :: [String] -> ProcessPars -> Summand -> Summand
fixGlobalsInSummand [] pars (params, c, a, aps, probChoices, g) = (params, c, a, aps, probChoices, g)
fixGlobalsInSummand (global:globals) pars (params, c, a, aps, probChoices, g) | length newValue > 1 = error("Fault in fixGlobalsInSummand")
                                                                              | otherwise           = result
  where
    indices      = [i | i <- [0..length pars - 1], fst (pars!!i) == global]
    firstIndex   = head indices
    newValue     = [g!!i | i <- indices, g!!i /= Variable global]
    realNewValue = if newValue == [] then Variable global else (newValue!!0)
    newG_        = [g!!i | i <- [0..length g - 1], not(elem i (tail indices))]
    newG         = [newG_!!i | i <- [0..firstIndex - 1]] ++ [realNewValue] ++ [newG_!!i | i <- [firstIndex + 1..length newG_ - 1]]
    result       = fixGlobalsInSummand globals [pars!!i | i <- [0..length pars -1], not (elem i (tail indices))] (params, c, a, aps, probChoices, newG)