----------------
--- SIMPLIFY ---
----------------

module Simplify where

import Auxiliary
import LPPE
import Expressions
import DataSpec
import List
import Processes

type TypeMapping    = [(Variable, Type)]

-- This function simplifies a specification by evaluating functions and removing 
-- unaccessible summands.
simplify :: Specification -> Specification
simplify (LPPE name pars summands, initialState, dataspec) = (LPPE name pars summandsNew, initialState, dataspec)
    where
       summands2   = map (simplifySummand dataspec) summands
       summandsNew = [summand | summand <- summands2, summandReachable dataspec (pars ++ getLocalPars summand) summand, not((getCondition summand) == Variable "F")]

summandReachable :: DataSpec -> TypeMapping -> Summand -> Bool
summandReachable dataspec mapping summand = isReachable
  where
    c           = getCondition summand
    isReachable = and [ nub [simplifyExpression dataspec (substituteInExpression [(par, Variable value)] c) | value <- getValues dataspec typ] /= [Variable "F"] | (par,typ) <- mapping]


simplifySummand :: DataSpec -> Summand -> Summand
simplifySummand dataspec (params, c, a, aps, probChoices, g)  = (params, c2, a, aps2, probChoices2, g2)
     where
       c2   = simplifyExpression dataspec c
       aps2 = map (simplifyExpression dataspec) aps
       g2   = map (simplifyExpression dataspec) g
       probChoices2 = simplifyProbChoices dataspec probChoices

simplifyProbChoices dataspec [] = []
simplifyProbChoices dataspec ((v,t,f):rest) = (v,t,f2):(simplifyProbChoices dataspec rest)
  where
    f2   = simplifyExpression dataspec f
    
-- This function simplifies an expression. 
simplifyExpression :: DataSpec -> Expression -> Expression
simplifyExpression dataspec (Variable v)                        = Variable v
simplifyExpression dataspec (Function "eq" [x,y]) | x == y      = Variable "T"
simplifyExpression dataspec (Function "and" [par])              = simplifyExpression dataspec par
simplifyExpression dataspec (Function "not" [Function "not" [par]]) = simplifyExpression dataspec par
simplifyExpression dataspec (Function "ifthenelse" [c, a, b])
  | isTrue                                                      = simplifyExpression dataspec a
  | isFalse                                                     = simplifyExpression dataspec b
  | simpleA == simpleB                                          = simpleA
  where
    isTrue     = simplifyExpression dataspec c == Variable "T"
    isFalse    = simplifyExpression dataspec c == Variable "F" 
    simpleA    = simplifyExpression dataspec a
    simpleB    = simplifyExpression dataspec b
simplifyExpression dataspec (Function "and" pars) 
  | elem (Variable "F") pars2                                   = Variable "F"
  | nonTrue == []                                               = Variable "T"
  | length nonTrue == 1                                         = nonTrue!!0
  | otherwise                                                   = Function "and" nonTrue
  where
    pars2   = map (simplifyExpression dataspec) pars
    nonTrue = [p | p <- pars2, p /= Variable "T"]
simplifyExpression dataspec (Function "or"  [par])              = simplifyExpression dataspec par
simplifyExpression dataspec (Function "or" pars) 
  | elem (Variable "T") pars2                                = Variable "T"
  | nonFalse == []                                              = Variable "F"
  | length nonFalse == 1                                        = nonFalse!!0
  | otherwise                                                   = Function "or" nonFalse
  where
    pars2    = map (simplifyExpression dataspec) pars
    nonFalse = [p | p <- pars2 , p /= Variable "F"]
simplifyExpression dataspec (Function name pars) | evaluatable  = Variable (evaluateFunction name (map constantExpressionToString simplePars) (thd3 dataspec))
                                                 | otherwise    = Function name simplePars
  where
    simplePars   = map (simplifyExpression dataspec) pars
    evaluatable  = and (map (expressionIsConstant dataspec) simplePars)

constantExpressionToString :: Expression -> [Char]
constantExpressionToString (Variable v) = v
constantExpressionToString other        = error("Function constantExpressionToString applied to non-constant.")
