module ToPA where

import Expressions
import LPPE
import List
import DataSpec 
import Processes
import Confluence
import Auxiliary
import Debug.Trace
import Data.HashMap
import Data.Maybe

type State             = [Constant]
type EdgeLabel         = String
type Probability       = String
type Transition        = (State, EdgeLabel, [(Probability, State)])
type TransitionInts    = (Int, EdgeLabel, [(Probability, Int)])
type Statespace        = ([State], [Transition]) 
type StatespaceInts    = ([State], [TransitionInts]) 
type Mapping           = [(Variable, Type, Constant)]
type Visited           = (Int, Int)
type RepresentationMap = HashMap State State

-------------------------------------------------
-- General functions to deal with state spaces --
-------------------------------------------------

-- Given a state space, provides the set of states
getStates :: StatespaceInts -> [State]
getStates (states, transitions) = states

-- Given a state space, provides the set of transitions
getTransitions :: StatespaceInts -> [TransitionInts]
getTransitions (states, transitions) = transitions

-------------------------------------------------
-- Functions for generation of the state space --
-------------------------------------------------

-- Given a system and a set of confluent summands, 
-- this function provides the number of states and transitions.
getStatesAndTransitionsConfluence :: Bool -> Specification -> ConfluentSummands -> Bool -> ((Int, Int), (Int, Int))
getStatesAndTransitionsConfluence ignorecycles spec confluent checkVisited = ((numberOfStates,numberOfTransitions), visited)
  where
    (statespace, initial, visited) = getStateSpace ignorecycles spec confluent checkVisited
    numberOfStates                 = length (getStates statespace)
    numberOfTransitions            = length (getTransitions statespace)

-- Given a system and a set of confluent summands,
-- this function provides the unfolded state space.
--
-- TODO: FIXEN DAT DE STATES DIE JE BEKEKEN HEBT VOOR INITIAL STATE OOK IN VISITED ERBIJ KOMEN.
getStateSpace :: Bool -> Specification -> ConfluentSummands -> Bool -> (StatespaceInts, State, (Int, Int))
getStateSpace ignorecycles spec confluent checkVisited = ((states, transitions), init, (fst visited2 + visitedStates, snd visited2 + visitedTrans))
  where
    (init, repr,visitedStates, visitedTrans) = getRepresentative spec (empty) confluent (getInitialState spec)
    ((states, transitions), visited2) = if not(checkVisited) then
	                                      (getStateSpace2 spec [init] 0 (singleton init 0, []) repr confluent, (0,0))
                                       else getStateSpace2Visited spec [init] 0 (singleton init 0, []) repr (0 ,0) confluent
    problem               = not(ignorecycles) && detectConfluentCycle spec confluent								          
--    warning               = if problem then "Warning: your model potentially contains a cycle of confluent summands.\n"
--	                        ++ "I will try to generate the state space anyway\n"
--	                        ++ "   (but might get stuck in an infinite loop).\n"
--	                        ++ "Alternatively, hide less actions, or use the -removecycles flag to resolve this\n"
--	                        ++ "   (but then you might get less reduction)." else ""

-- This function takes an LPPE, a set of states to explore, the state space
-- that was generated so far, and a set of confluent summands.
getStateSpace2 :: Specification -> [State] -> Int -> (HashMap State Int, [TransitionInts]) -> RepresentationMap -> ConfluentSummands -> StatespaceInts
getStateSpace2 spec []             highest statespace repr confluent = (keys (fst statespace), snd statespace)
getStateSpace2 spec (state:states) highest statespace repr confluent
  | forceEvaluation = error("")
  | otherwise       = getStateSpace2 spec toExplore highest2 updatedStatespace newRepr confluent
  where
    ((newStates, newTransitions), 
      newRepr, _)                  = exploreState spec state state repr confluent False
    (statesSeen, transitionsSeen)  = statespace
    reallyNewStates                = findReallyNewStates (nub newStates) statesSeen
    reallyNewTransitions           = nub newTransitions
    (updatedstates, highest2)      = addListToMap reallyNewStates highest statesSeen
    optimisedTransitions           = List.map (changeTransition updatedstates) reallyNewTransitions
    updatedtransitions             = (optimisedTransitions ++ transitionsSeen)
    updatedStatespace              = (updatedstates, updatedtransitions)
    toExplore                      = (reallyNewStates ++ states)
    forceEvaluation                = [0 | (s,a,t) <- optimisedTransitions, s < -1 || length a < -1 || or [length a < -1 || b < -1 | (a,b) <- t]] /= []

-- This function takes an LPPE, a set of states to explore, the state space
-- that was generated so far, and a set of confluent summands.
getStateSpace2Visited :: Specification -> [State] -> Int -> (HashMap State Int, [TransitionInts]) -> RepresentationMap -> Visited -> ConfluentSummands -> (StatespaceInts, (Int, Int))
getStateSpace2Visited spec []             highest statespace repr visited confluent = ((keys (fst statespace), snd statespace), visited)
getStateSpace2Visited spec (state:states) highest statespace repr visited confluent
  | forceEvaluation = error("")
  | otherwise       = getStateSpace2Visited spec toExplore highest2 updatedStatespace newRepr newVisited2 confluent
  where
    ((newStates, newTransitions), 
     newRepr, newVisited1)         = exploreState spec state state repr confluent True
    (statesSeen, transitionsSeen)  = statespace
    reallyNewStates                = findReallyNewStates (nub newStates) statesSeen
    reallyNewTransitions           = nub newTransitions
    (updatedstates, highest2)      = addListToMap reallyNewStates highest statesSeen
    optimisedTransitions           = List.map (changeTransition updatedstates) reallyNewTransitions
    updatedtransitions             = (optimisedTransitions ++ transitionsSeen)
    updatedStatespace              = (updatedstates, updatedtransitions)
    toExplore                      = (reallyNewStates ++ states)
    visitedTransitions             = snd newVisited1 + snd visited
    visitedStates                  = fst newVisited1 + fst visited --addListToMap2 updatedstates (fst newVisited1) (fst visited)
    newVisited2                    = (visitedStates, visitedTransitions)
    forceEvaluation                = visitedStates < -1 || visitedTransitions < -1 || [0 | (s,a,t) <- optimisedTransitions, s < -1 || length a < -1 || or [length a < -1 || b < -1 | (a,b) <- t]] /= []

changeTransition :: HashMap State Int -> Transition -> TransitionInts
changeTransition states (i, a, next) = (fromJust (Data.HashMap.lookup i states), a, 
                                       List.map (\x -> (fst x, fromJust (Data.HashMap.lookup (snd x) states))) next)

findReallyNewStates :: [State] -> HashMap State Int -> [State]
findReallyNewStates []     seen = []
findReallyNewStates (x:xs) seen | member x seen = findReallyNewStates xs seen
                                | otherwise     = x:(findReallyNewStates xs seen)

addListToMap :: [State] -> Int -> HashMap State Int -> (HashMap State Int, Int)
addListToMap [] highest seen     = (seen, highest)
addListToMap (x:xs) highest seen = addListToMap xs (highest + 1) (Data.HashMap.insert x (highest + 1) seen)

addListToMap2 :: HashMap State Int -> [State] -> HashMap State Int -> HashMap State Int
addListToMap2 seen [] visited     = visited
addListToMap2 seen (x:xs) visited | member x seen || member x visited = addListToMap2 seen xs visited
                                  | otherwise                         = addListToMap2 seen xs (Data.HashMap.insert x 0 visited)

-- sourceState geeft aan welke state als source-state in de transities moet komen te staan.
-- Als je ctau-stappen hebt dan ga je de toestand waar die uitkomt zelf niet in de statespace zetten, 
-- maar de concatenatie van de transitie van sourceState naar from en de transities vanuit from
exploreState :: Specification -> State -> State -> RepresentationMap -> ConfluentSummands -> Bool -> (Statespace, RepresentationMap, (Int, Int))
exploreState spec from sourceState repr confluent countVisited | not countVisited = (newTransitions, newRepr, (0,0)) 
                                                               | otherwise        = (newTransitions, newRepr, (visitedStates,visitedTransitions + visitedTrans))
  where
    (LPPE name params summands)     = getLPPE spec
    mapping                         = zip3 (List.map fst params) (List.map snd params) from
    trans_                          = potentialBehaviour False spec confluent (getDataSpec spec) sourceState mapping [summands!!i | i <- [0.. length summands - 1], not (elem i confluent)]
    visitedTransitions              = length trans_
    (trans, newRepr, 
     visitedStates1, visitedTrans)  = if confluent /= [] then changeToRepresentatives spec repr confluent trans_
	                                                     else (trans_, repr, 0, 0)
    states                          = concat [List.map snd list | (s,a,list) <- trans]
    newTransitions                  = (states, trans)
    visitedStates                   = visitedStates1 -- + 1 --states ++ (concat [List.map snd list | (s,a,list) <- trans_]) ++ visitedStates1


changeToRepresentatives :: Specification -> RepresentationMap -> ConfluentSummands -> [Transition] -> ([Transition], RepresentationMap, Int, Int)
changeToRepresentatives _ repr _ [] = ([], repr, 0, 0)
changeToRepresentatives spec repr confluent (t:rest) = (newT:newRest, newRepr2, visited1 + visited2, visitedTrans1 + visitedTrans2)
  where
    (newT, newRepr1,visited1,visitedTrans1) = changeTransitionToRepresentatives spec repr confluent t
    (newRest, newRepr2,visited2, visitedTrans2) = changeToRepresentatives spec newRepr1 confluent rest

changeTransitionToRepresentatives :: Specification -> RepresentationMap -> ConfluentSummands -> Transition -> (Transition, RepresentationMap, Int, Int)
changeTransitionToRepresentatives spec repr confluent (s,a,next) = ((newS, a, newNext), newRepr2, visited, visitedTrans)
  where
    newS     = s
    (newNext, newRepr2, visited, visitedTrans) = changeNextStateToRepresentatives spec repr confluent next   

changeNextStateToRepresentatives :: Specification -> RepresentationMap -> ConfluentSummands -> [(Probability, State)] -> ([(Probability, State)], RepresentationMap, Int, Int)
changeNextStateToRepresentatives spec repr confluent [] = ([], repr, 0, 0)
changeNextStateToRepresentatives spec repr confluent ((p,s):rest) = ((p,newS):newRest, newRepr2, visited1 + visited2, visitedTrans1 + visitedTrans2)
  where
    (newS, newRepr1, visited1, visitedTrans1)    = getRepresentative spec repr confluent s
    (newRest, newRepr2, visited2, visitedTrans2) = changeNextStateToRepresentatives spec newRepr1 confluent rest

potentialBehaviour a b c d e f g = potentialBehaviour2 a b c d e f g

potentialBehaviour2 :: Bool -> Specification -> ConfluentSummands -> DataSpec -> State -> Mapping -> [Summand] -> [Transition]
--potentialBehaviour followConfluent spec confluents dataspec sourceState mapping _      | trace(show followConfluent ++ " " ++ show sourceState) False = error("Blaat") 
potentialBehaviour2 followConfluent spec confluents dataspec sourceState mapping []     = []
potentialBehaviour2 followConfluent spec confluents dataspec sourceState mapping (s:ss) = trans ++ rest
  where
    trans = exploreSummand followConfluent spec confluents dataspec sourceState mapping s
    rest  = potentialBehaviour2 followConfluent spec confluents dataspec sourceState mapping ss

exploreSummand :: Bool -> Specification -> ConfluentSummands -> DataSpec -> State -> Mapping -> Summand -> [Transition]
exploreSummand followConfluent spec confluent dataspec sourceState mapping (((var,sort):params), c, a, aps, probChoices, g)
  = unfoldSummation followConfluent spec confluent dataspec sourceState mapping (var,sort) (getValues dataspec sort) (params, c, a, aps, probChoices, g)
exploreSummand followConfluent spec confluent dataspec sourceState mapping ([], c, a, aps, probChoices, g)
  | executable = transitions -- (states, transitions)
  | otherwise  = [] -- ([], [])
  where
    executable  = evalExpr dataspec mapping c == "T"
    steps       = generateTransitions followConfluent spec confluent dataspec mapping probChoices g "1"
--    states      = List.map snd steps
    transitions = [(sourceState,a ++ (optionalParentheses (infixString (List.map (evalExpr dataspec mapping) aps) ",")), steps)]

unfoldSummation :: Bool -> Specification -> ConfluentSummands -> DataSpec -> State -> Mapping -> (Variable,Type) -> [Constant] -> Summand -> [Transition]
unfoldSummation followConfluent spec confluent dataspec sourceState mapping (var,sort) [] summand = []
unfoldSummation followConfluent spec confluent dataspec sourceState mapping (var,sort) (value:values) summand = t1 ++ t2
  where
    mapping2 = updateMapping mapping (var,sort) value
    t1       = exploreSummand followConfluent spec confluent dataspec sourceState mapping2 summand
    t2       = unfoldSummation followConfluent spec confluent dataspec sourceState mapping (var,sort) values summand

generateTransitions :: Bool -> Specification -> ConfluentSummands -> DataSpec -> Mapping -> [(Variable, Type, ProbDistr)] -> NextPars -> String -> [(Probability, State)]
generateTransitions followConfluent spec confluent dataspec mapping [] g frac | nextProb /= "0" = [(nextProb, nextState2)]
                                                                                 | otherwise       = []
  where
    nextState  = List.map (evalExpr dataspec mapping) g
    nextState2 = nextState -- if followConfluent then nextState else getRepresentative spec confluent nextState
    nextProb   = frac
generateTransitions followConfluent spec confluent dataspec mapping ((v,t,f):probChoices) g frac
  = unfoldProbChoice followConfluent spec confluent dataspec mapping (v,t) f g (getValues dataspec t) probChoices frac
   
unfoldProbChoice :: Bool -> Specification -> ConfluentSummands -> DataSpec -> Mapping -> (Variable, Type) -> ProbDistr -> NextPars -> [Constant] -> [(Variable, Type, ProbDistr)] -> String -> [(Probability, State)]
unfoldProbChoice followConfluent spec confluent dataspec mapping (v,t) f g [] probChoices frac             = []
unfoldProbChoice followConfluent spec confluent dataspec mapping (v,t) f g (value:values) probChoices frac = steps1 ++ steps2
  where
    mapping2 = updateMapping mapping (v,t) value
    nextProb = evalExpr dataspec mapping2 f
    newFrac  = writeFraction ((getFraction frac) * (getFraction nextProb))
    steps1   = generateTransitions followConfluent spec confluent dataspec mapping2 probChoices g newFrac
    steps2   = unfoldProbChoice followConfluent spec confluent dataspec mapping (v,t) f g values probChoices frac

evalExpr :: DataSpec -> Mapping -> Expression -> Constant
evalExpr dataspec mapping (Function name pars)              = evaluateFunction name (List.map (evalExpr dataspec mapping) pars) (thd3 dataspec)
evalExpr dataspec mapping (Variable v) | value == Nothing   = v
                                       | otherwise          = Data.Maybe.fromJust value
  where
    value = getValue mapping v

getValue :: Mapping -> [Char] -> Maybe Constant
getValue [] k = Nothing -- error("No value defined for variable " ++ k)
getValue ((key,sort,value):mapping) k | key == k = Just value
                                      | otherwise = getValue mapping k

updateMapping :: Mapping -> (String, Type) -> Constant -> Mapping
updateMapping [] (v,t) newVal = [(v, t, newVal)]
updateMapping ((key,sort,value):mapping) (v,t) newVal | key == v  = ((v,t,newVal):mapping)
                                                  | otherwise = ((key,sort,value):(updateMapping mapping (v,t) newVal))

---------------------
-- Representatives --
---------------------

type NumbersTable = [(State, Int)]
type ListTable    = [(State, [State])]
type StateTable  = [(State, State)]

getRepresentative spec repr confluent state | representative /= Nothing = (fromJust representative, repr, 0, 0) 
                                            | otherwise = (newRepresentative, newRepresentatives, length visited, visitedTrans)
  where
    numbers                                   = setItem [] state 0
    representative                            = Data.HashMap.lookup state repr
    (newRepresentative,visited, visitedTrans) = getRepresentative2 repr spec confluent state numbers [] [] [] 0
 -- DO NOT STORE REPRESENTATIVES:   newRepresentatives                         = repr
 -- ONLY STORE STATES THAT ARE A REPRESENTATIVE THEMSELVES:  newRepresentatives                       = Data.HashMap.insert newRepresentative newRepresentative repr
    newRepresentatives                       = Data.HashMap.insert state newRepresentative repr  -- ONLY STORE REPRESENTATIVES FOR STATES THAT HAVE BEEN ASKED FOR THEIR REPRESENTATIVE
--  STORE EVERYTHING:  newRepresentatives_                       = Data.HashMap.insert newRepresentative newRepresentative repr
 --   newRepresentatives                        = Data.HashMap.insert newRepresentative newRepresentative (addBlaat newRepresentatives_ newRepresentative visited)
 --   nrVisited                                 = if newRepresentative == state then 0 else length visited

addBlaat newRepresentatives representative [] = newRepresentatives
addBlaat newRepresentatives representative (s:ss) = addBlaat (Data.HashMap.insert s representative  newRepresentatives) representative ss 

getRepresentative2 repr spec confluent state numbers low unexplored backtrack count 
  | already   = (existingRepresentative, visited1, visitedTrans1)
  | found     = (findLowestRepresentative state low, visited1, visitedTrans1)
  | otherwise = (rep, visited1 ++ visited2, visitedTrans1 + visitedTrans2)
  where
    newState      = getNumberFromList numbers state == 0
    (numbers2, low2, unexplored2, count2, visited1, already, existingRepresentative, visitedTrans1) 
             = if newState then exploreNewState repr spec confluent state numbers low unexplored count
                           else (numbers, low, unexplored, count, [], False, [], 0)
    deadlockState = getItem unexplored2 state == []
    (found, low3, state2)                          = if deadlockState      then exploreDeadlockState state numbers2 unexplored2 low2 backtrack  
                 	                                                       else (False, low2, state)
    (unexplored3, backtrack2, state3, low4)        = if not(deadlockState) then getRepresentative3 unexplored2 numbers2 backtrack state low2
	                                                                       else (unexplored2, backtrack, state2, low3)
    (rep, visited2, visitedTrans2) = getRepresentative2 repr spec confluent state3 numbers2 low4 unexplored3 backtrack2 count2

exploreNewState :: RepresentationMap -> Specification -> ConfluentSummands -> State -> NumbersTable -> NumbersTable -> ListTable -> Int -> (NumbersTable, NumbersTable, ListTable, Int, [State], Bool, State, Int)
exploreNewState repr spec confluents v numbers low unexplored count = (numbers3, low2, unexplored3, count2, [v], already, existing, visitedTransitions)
  where
    count2                      = count + 1
    numbers2                    = setItem numbers v count2
    low2                        = setItem low v count2
    unexplored2                 = setItem unexplored v []
    (LPPE name params summands) = getLPPE spec
    mapping                     = zip3 (List.map fst params) (List.map snd params) v
    trans                       = potentialBehaviour True spec confluents (getDataSpec spec) v mapping [summands!!i | i <- [0.. length summands - 1], elem i confluents]
    newstates                   = concat [List.map snd next | (s,a,next) <- trans]
    (unexplored3, numbers3, already, existing) = addTransitions repr v unexplored2 numbers2 newstates
    visitedTransitions          = length trans

addTransitions :: RepresentationMap -> State -> ListTable -> NumbersTable -> [State] -> (ListTable, NumbersTable, Bool, State)
addTransitions repr v unexplored numbers []     = (unexplored, numbers, False, [])
addTransitions repr v unexplored numbers (t:ts) | existing /= Nothing = (unexplored, numbers, True, fromJust existing)
                                                | otherwise = addTransitions repr v unexplored2 numbers2 ts
  where
    unexplored2 = setItem unexplored v (t:(getItem unexplored v))
    numbers2    = if getNumberFromList numbers t == -1 then setItem numbers t 0 else numbers
    existing    = Data.HashMap.lookup t repr

exploreDeadlockState :: State -> NumbersTable -> ListTable -> NumbersTable -> StateTable -> (Bool, NumbersTable, State)
exploreDeadlockState v numbers unexplored low backtrack | found     = (True, [], [])
                                                        | otherwise = (False, low2, v2)
  where
    lowV  = getNumberFromList low v
    found = getNumberFromList numbers v == lowV 
    prev  = getItem backtrack v
    low2  = setItem low prev (min (getNumberFromList low prev) lowV)
    v2    = prev

getRepresentative3 :: ListTable -> NumbersTable -> StateTable -> State -> NumbersTable -> (ListTable, StateTable, State, NumbersTable)
getRepresentative3 unexplored numbers backtrack v low = (unexplored2, backtrack2, v2, low2)
  where
    rest        = getItem unexplored v
    u           = head rest
    unexplored2 = setItem unexplored v (tail rest)
    nrU         = getNumberFromList numbers u
    dosomething = nrU == 0
    backtrack2  = if dosomething then setItem backtrack u v else backtrack
    v2          = if dosomething then u else v
    low2        = if not(dosomething) && nrU < (getNumberFromList numbers v) then setItem low v (min (getNumberFromList low v) (getNumberFromList low u)) else low

findLowestRepresentative v low = foldl min v options
  where
    options = [state | (state, nr) <- low, nr == getNumberFromList low v]

setItem :: [(State, a)] -> State -> a -> [(State, a)]
setItem [] v i                           = [(v,i)]
setItem ((val, nr):rest) v i | val == v  = (val, i):rest
                               | otherwise = (val,nr):(setItem rest v i)

getItem :: [(State, a)] -> State -> a
getItem ((v,i):rest) from | from == v = i
                            | otherwise = getItem rest from

getNumberFromList :: NumbersTable -> State -> Int
getNumberFromList [] s = -1
getNumberFromList ((v,i):rest) from | from == v = i
                                    | otherwise = getNumberFromList rest from